using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Markup;
using System.Windows.Media;
using System.Windows.Controls;
using System.Windows.Media.Imaging;
using System.Windows;
using System.Diagnostics;

namespace MidPointFills
{
    /// <summary>
    /// Markup extension offering midpoint interpolated linear gradient fills.
    /// </summary>
    /// <remarks>
    /// WPF's LinearGradientBrush only performs linear interpolation between
    /// gradient stops. (That might sounds like a "well duh" kind of observation,
    /// but actually, the 'Linear' in LinearGradientBrush refers to the fact that
    /// the gradient itself runs in a straight line.)
    /// 
    /// This class offers linear gradients with support for midpoints. These are
    /// a common feature of gradient fills - Adobe Illustrator has them, for example -
    /// enabling a specific form of non-linear interpolation to be used. If a
    /// midpoint is specified between two gradient stops, and if that midpoint is
    /// anything other than dead-centered between the stops, the interpolation
    /// fits a curve that arranges for the 50% point to be where the midpoint lies.
    /// 
    /// The curve is an exponential curve. The start and end points of the curve
    /// are determined by the gradient stops on either side. The curve is made to
    /// fit by adjusting the exponential power so that the midpoint falls where
    /// specified.
    /// 
    /// The lower end of the curve is always in the larger of the two pieces into
    /// which the midpoint divides the fill segment. This means that the direction
    /// of the curve flips as the midpoint crosses the half way point. (This doesn't
    /// cause a discontinuity, because the curve is a straight line if the midpoint
    /// is at the halfway point.) This is done for consistency with Illustrator.
    /// Illustrator presumably does this because it results in symmetric-looking
    /// fills when the midpoints themselves are arranged symmetrically on either
    /// side of a stop.
    /// 
    /// This markup extension returns an ImageBrush. We cannot add a custom fill
    /// type to WPF, so we have to construct a bitmap that approximates the
    /// fill we would like.
    /// </remarks>
    [ContentProperty("GradientsAndMidpoints")]
    public class LinearGradientWithMidpointsBrushExtension : MarkupExtension
    {
        /// <summary>
        /// The gradient stops and optional midpoints describing the fill.
        /// </summary>
        /// <remarks>
        /// Users of this element would normally set this property from
        /// XAML. Is is the content property, so elements added inside this
        /// LinearGradientWithMidPointsBrushExtension will automatically
        /// be added to this property.
        /// 
        /// Note that this is defined as a collection of objects, since it
        /// accepts more than one type. In fact, it only accepts GradientStop
        /// and GradientMidpoint objects. Moreover, the first item must be a
        /// GradientStop, and there must be either 0 or 1 GradientMidpoints
        /// between each GradientStop.</remarks>
        public List<object> GradientsAndMidpoints
        {
            get { return gradientsAndMidpointsValue; }
        }
        List<object> gradientsAndMidpointsValue = new List<object>();


        /// <summary>
        /// Markup extension value provider method.
        /// </summary>
        /// <param name="serviceProvider">Context provided by XAML parser.</param>
        /// <returns>An ImageBrush representing the specified fill.</returns>
        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            int[] pixelValues = GeneratePixelValues();

            WriteableBitmap bmp = new WriteableBitmap(pixelValues.Length, 1, 96, 96, PixelFormats.Pbgra32, null);
            bmp.WritePixels(new Int32Rect(0, 0, pixelValues.Length, 1), pixelValues, pixelValues.Length, 0);

            // This is a bit of a hack - we're ignoring the StartPoint and EndPoint of the fill.
            ImageBrush brush = new ImageBrush(bmp);

            return brush;
        }

        /// <summary>
        /// Information about a segment that makes up the fill. 
        /// </summary>
        /// <remarks>
        /// The API we present is designed to be easy for consumers of
        /// this class - they can pass in a set of GradientStop elements
        /// just like they can with the built in gradient fills, and they
        /// can also intermingle GradientMidpoints where they want them.
        /// (A central midpoint is presumed when none is specified.)
        /// But that's not an especially convenient representation when
        /// it comes to building the fill. It's easy to work in terms of
        /// a series of segments. (I.e., the bits between the stops, rather
        /// than the stops themselves.) So we build a collection of these
        /// FillSegment structures to make life easier.
        /// </remarks>
        private class FillSegment
        {
            public Color start;
            public double startpoint;
            public double midpoint;
            public double endpoint;
            public Color end;
        }


        /// <summary>
        /// Builds a description of the segements of the fill from the current
        /// value of GradiensAndMidpoints.
        /// </summary>
        /// <returns>A colelction of FillSegments describing the fill.</returns>
        private List<FillSegment> BuildSegmentList()
        {
            // Demo code only... Fails to deal with out-of-order
            // stops, or missing stops.

            List<FillSegment> segments;
            segments = new List<FillSegment>();
            bool hadMidpoint = false;
            foreach (object item in GradientsAndMidpoints)
            {
                GradientStop stop = item as GradientStop;
                if (stop != null)
                {
                    if (segments.Count != 0)
                    {
                        // Finish off the prevous segment.
                        segments[segments.Count - 1].endpoint = stop.Offset;
                        segments[segments.Count - 1].end = stop.Color;

                        // If no midpoint was specified, put it dead centre
                        if (!hadMidpoint)
                        {
                            segments[segments.Count - 1].midpoint = (segments[segments.Count - 1].startpoint +
                                segments[segments.Count - 1].endpoint) / 2;
                        }
                    }

                    FillSegment newSegment = new FillSegment();
                    newSegment.start = stop.Color;
                    newSegment.startpoint = stop.Offset;
                    segments.Add(newSegment);
                    hadMidpoint = false;
                }
                else
                {
                    GradientMidpoint midPoint = item as GradientMidpoint;
                    if (midPoint == null)
                    {
                        throw new ArgumentException("GradientsAndMidpoints collection must contain only GradientStop or GradientMidpoint objects");
                    }
                    if (segments.Count == 0)
                    {
                        throw new ArgumentException("First item in GradientsAndMidpoints collection must be a GradientStop");
                    }
                    if (hadMidpoint)
                    {
                        throw new ArgumentException("GradientsAndMidpoints collection can contain at most one GradientMidpoint for each GradientStop");
                    }
                    segments[segments.Count - 1].midpoint = midPoint.Offset;
                    hadMidpoint = true;
                }
            }
            // The loop above ends up generating a spurious extra final segment.
            segments.RemoveAt(segments.Count - 1);
            return segments;
        }


        /// <summary>
        /// Creates an array of PBGRA-format pixel values to represent the
        /// fill.
        /// </summary>
        /// <returns>An array containing pixel values that represent the
        /// fill.</returns>
        private int[] GeneratePixelValues()
        {
            List<FillSegment> segments = BuildSegmentList();

            int pixelCount = CalculateGradientPixelCount(segments);
            int[] pixelValues = new int[pixelCount];
            for (int i = 0; i < pixelCount; ++i)
            {
                double offset = ((double) i) / ((double) pixelCount);
                foreach (FillSegment seg in segments)
                {
                    if (offset >= seg.startpoint && offset <= seg.endpoint)
                    {
                        double segOffset = (offset - seg.startpoint) / (seg.endpoint - seg.startpoint);
                        double pm = (seg.midpoint - seg.startpoint) / (seg.endpoint - seg.startpoint);

                        bool flip = false;
                        if (pm < 0.5)
                        {
                            pm = 1 - pm;
                            segOffset = 1 - segOffset;
                            flip = true;
                        }
                        double power = Math.Log(0.5) / Math.Log(pm);
                        double scale = Math.Pow(segOffset, power);
                        if (flip) { scale = 1 - scale; }

                        byte a = (byte) ((1 - scale) * seg.start.A + scale * seg.end.A);
                        byte r = (byte) ((1 - scale) * seg.start.R + scale * seg.end.R);
                        byte g = (byte) ((1 - scale) * seg.start.G + scale * seg.end.G);
                        byte b = (byte) ((1 - scale) * seg.start.B + scale * seg.end.B);

                        pixelValues[i] = (a << 24) + (r << 16) + (g << 8) + b;
                        break;
                    }
                }

            }
            return pixelValues;
        }

        /// <summary>
        /// Works out how many pixels are required to represent the fill reasonably
        /// accurately.
        /// </summary>
        /// <param name="segments">A description of the fill.</param>
        /// <returns>The number of pixels required to represent this fill.</returns>
        private static int CalculateGradientPixelCount(List<FillSegment> segments)
        {
            int pixelCount;
            // The bitmap for our image brush needs to be big enough so that
            // the part of the curve with the highest rate of change doesn't
            // band.
            // Since this is demo code, we hack it, and just go on the shortest
            // span size (where a 'span' is distance between a midpoint and a
            // gradient stop). Really, we should be looking at the rate of
            // change...

            double shortestSpan = double.PositiveInfinity;
            foreach (FillSegment seg in segments)
            {
                double first = seg.midpoint - seg.startpoint;
                double second = seg.endpoint - seg.startpoint;
                shortestSpan = Math.Min(shortestSpan, Math.Min(first, second));
            }

            pixelCount = (int) (256.0 / shortestSpan);
            return pixelCount;
        }
    }
}
